﻿namespace XmobiTea.MiniPoolManager
{
    using XmobiTea.MiniSingleton;
    using UnityEngine;
    using XmobiTea.MiniGlobalCoroutine;
    using System.Collections.Generic;

    public class Pool
    {
        public Stack<GameObject> DeactiveStack { get; private set; }
        public Pool()
        {
            DeactiveStack = new Stack<GameObject>();
        }
    }

    public class PoolManager : Singleton<PoolManager>
    {
        private Dictionary<string, Pool> PoolDic { get; set; }
        private Dictionary<string, Transform> PoolParentDic { get; set; }

        protected override void OnInit()
        {
            base.OnInit();

            if (PoolDic == null) PoolDic = new Dictionary<string, Pool>();
            PoolDic.Clear();

            if (PoolParentDic == null) PoolParentDic = new Dictionary<string, Transform>();
            PoolParentDic.Clear();
        }

        public static GameObject Spawn(GameObject prefab, Vector3 position, Quaternion rotation)
        {
            if (!IsExists)
            {
                Debug.LogError("[Pool] PoolManager instance null");
                return null;
            }

            if (prefab == null)
            {
                Debug.LogError("[Pool] Can not instantiate a null prefab");
                return null;
            }

            var iPoolItem = prefab.GetComponent<IPoolItem>();
            if (iPoolItem == null)
            {
                iPoolItem = prefab.AddComponent<PoolItem>();
                iPoolItem.PoolName = prefab.name + '-' + System.Guid.NewGuid().ToString("N");
            }

            var poolName = iPoolItem.PoolName;

            var poolDic = Instance.PoolDic;

            var pool = poolDic.ContainsKey(poolName) ? poolDic[poolName] : null;

            if (pool == null)
            {
                pool = new Pool();
                poolDic[poolName] = pool;
            }

            var deactiveStack = pool.DeactiveStack;

            GameObject instance;

            if (deactiveStack.Count != 0)
            {
                instance = deactiveStack.Pop();

                instance.transform.position = position;
                instance.transform.rotation = rotation;
            }
            else
            {
                var poolParentDic = Instance.PoolParentDic;

                var parent = poolParentDic.ContainsKey(poolName) ? poolParentDic[poolName] : null;
                if (parent == null)
                {
                    parent = new GameObject(poolName).transform;
                    parent.parent = Instance.transform;

                    poolParentDic[poolName] = parent;
                }

                instance = Instantiate(prefab, position, rotation);
                instance.transform.parent = parent;
            }

            if (!instance.activeSelf) instance.SetActive(true);
            instance.GetComponent<IPoolItem>().OnSpawn();

            return instance;
        }

        public static void Despawn(GameObject instance, float timeToLive = 0)
        {
            if (!IsExists)
            {
                Debug.LogError("[Pool] PoolManager instance null");
                return;
            }

            if (instance == null)
            {
                Debug.LogError("[Pool] Can not get null instance");
                return;
            }

            if (timeToLive != 0 && !float.IsNaN(timeToLive))
            {
                GlobalCoroutine.RunCoroutine(Instance.IEDespawn(instance, timeToLive));
                return;
            }

            var iPoolItem = instance.GetComponent<IPoolItem>();

            if (iPoolItem == null)
            {
                Debug.LogError("[Pool] iPoolItem not exists in this instance");
                return;
            }

            var poolName = iPoolItem.PoolName;

            var poolDic = Instance.PoolDic;

            var pool = poolDic.ContainsKey(poolName) ? poolDic[poolName] : null;

            if (pool == null)
            {
                Debug.LogError("[Pool] pool for this instance is null");
                return;
            }

            if (instance.activeSelf)
            {
                instance.SetActive(false);
                pool.DeactiveStack.Push(instance);

                iPoolItem.OnDespawn();
            }
        }

        private System.Collections.IEnumerator IEDespawn(GameObject instance, float timeToLive)
        {
            var timer = Time.time + timeToLive;

            yield return new WaitWhile(() => timer > Time.time && instance != null && instance.activeSelf);

            if (instance != null && instance.activeSelf)
                Despawn(instance);
        }
    }
}
